"""Submission discovery and loading from the registrar prefix."""

import importlib
import importlib.util
import pathlib
import sys
from dataclasses import dataclass
from importlib.machinery import ModuleSpec
from types import ModuleType

from sneks.engine.config.base import Config
from sneks.engine.core.snek import Snek


@dataclass(frozen=True)
class Submission:
    """A discovered submission with name and Snek class."""

    name: str
    snek: type[Snek]


def get_submissions() -> list[Submission]:
    sneks: list[Submission] = []
    snek_classes = get_submission_classes()
    for name, snek in snek_classes.items():
        if Config().registrar_submission_sneks > 1:
            for i in range(Config().registrar_submission_sneks):
                sneks.append(Submission(f"{name}{i}", snek))  # type: ignore
        else:
            sneks.append(Submission(name, snek))  # type: ignore
    return sneks


def get_submission_classes() -> dict[str, type[Snek]]:
    results = {}
    submissions = set(
        p.parent
        for p in pathlib.Path(Config().registrar_prefix).glob("**/submission.py")
    )
    for submission in submissions:
        name, snek = get_custom_snek(submission)
        if name is not None and snek is not None:
            results[name] = snek
    return results


def get_custom_snek(prefix: pathlib.Path) -> tuple[str | None, type[Snek] | None]:
    name, module = load_module(prefix)
    if module is None:
        return None, None
    return name, module.CustomSnek


def load_module(prefix: pathlib.Path) -> tuple[str | None, ModuleType | None]:
    name, spec, module = get_module(prefix)
    if name is not None and spec is not None and module is not None:
        assert spec.loader is not None
        spec.loader.exec_module(module)
        sys.modules[name] = module
    return name, module


def get_module(
    prefix: pathlib.Path,
) -> tuple[str | None, ModuleSpec | None, ModuleType | None]:
    submission = get_submission(prefix)
    if submission is None:
        return None, None, None
    name = get_submission_name(submission)
    spec = importlib.util.spec_from_file_location(name, submission)
    module = importlib.util.module_from_spec(spec) if spec is not None else None
    return name, spec, module


def get_submission(prefix: pathlib.Path) -> pathlib.Path | None:
    files = get_submission_files(prefix)
    if files:
        return files[0]
    return None


def get_submission_files(prefix: pathlib.Path) -> list[pathlib.Path]:
    suffix = "submission.py"
    return list(prefix.glob(f"**/{suffix}"))


def get_submission_name(prefix: pathlib.Path) -> str:
    return str(prefix.relative_to(Config().registrar_prefix).parent)



"""Base configuration dataclass for game settings."""

from dataclasses import dataclass, field

from sneks.engine.config.entity import EntityConfig
from sneks.engine.config.graphics import GraphicsConfig
from sneks.engine.config.snek import SnekConfig
from sneks.engine.config.world import WorldConfig
from sneks.engine.util import Singleton


@dataclass(frozen=True)
class Config(metaclass=Singleton):
    title: str = "Sneks"
    world: WorldConfig = field(default_factory=WorldConfig)
    entity: EntityConfig = field(default_factory=EntityConfig)
    snek: SnekConfig = field(default_factory=SnekConfig)
    graphics: GraphicsConfig = field(default_factory=GraphicsConfig)
    runs: int = 1
    turn_limit: int = 1000
    registrar_prefix: str = "src"
    registrar_submission_sneks: int = 1
    profile: bool = False
    debug: bool = True



"""DynamicEntity for entities that move and change shape."""

from abc import ABC
from dataclasses import dataclass

from typing_extensions import override

from sneks.engine.core.cell import Cell
from sneks.engine.entity.base import Entity
from sneks.engine.manager.cell import CellManager


@dataclass(kw_only=True)
class DynamicEntity(Entity, ABC):
    """An entity whose shape changes."""

    @property
    def head(self) -> Cell:
        return self.position_cells[0]

    @override
    def step(self, idle_allowed: bool = True) -> None:
        """Default behavior for dynamic entities.

        The default is to have the collision box trail behind the head
        based on the speed at which the entity is traveling. This is to prevent
        the snek from phasing through other entities at higher speeds.

        The decorative cells will contain a history of where the position cells
        have traveled previously. How they're rendered is delegated to the
        graphics configuration.
        """

        # we need to track the colliding cells from the head separate from the tail
        # because we'll use the head colliding cells to tell if this entity has
        # collided into another entity, and the tail colliding cells to determine
        # what is able to be collided into

        # defer to super to adjust bearing
        super().step(idle_allowed=idle_allowed)

        # update the cells
        components = self.bearing_manager.get_direction_components()
        component_count = 0
        for direction in components:
            component_count += 1
            next_head = CellManager.get_absolute_neighbor(self.head, direction)
            self.position_cells.appendleft(next_head)
            self.head_colliding_cells.appendleft(next_head)

        # trim the position cells to the size, appending to decorative and tail colliding
        while len(self.position_cells) > self.size:
            cell = self.position_cells.pop()
            self.tail_colliding_cells.appendleft(cell)
            self.decorative_cells.appendleft(cell)

        # trim the colliding cells to the speed
        while len(self.head_colliding_cells) > max(0, component_count - 1):
            self.head_colliding_cells.pop()
            if self.tail_colliding_cells:
                self.tail_colliding_cells.pop()

        # trim the decorative cells based on a factor of size * speed
        while len(self.decorative_cells) > 3 * self.size * component_count:
            self.decorative_cells.pop()

        # example movement
        # moving left at speed 2
        # - - - - - x
        # - - - x + -
        # - x + - - -
        # or size bigger
        # - - - - - X x x
        # - - - X x x + -
        # - X x x + - - -



"""Frame capture and video encoding for game recording."""

import os
import pathlib
import subprocess
import uuid

import pygame.image
from pygame import Surface

from sneks.engine.config.base import Config

os.environ["PYGAME_HIDE_SUPPORT_PROMPT"] = "1"


class Recorder:
    """Recorder for capturing frames and encoding video."""

    def __init__(self) -> None:
        if not Config().graphics.record:
            return

        self.identifier = uuid.uuid4()
        self.i = 0
        self.prefix = pathlib.Path(Config().graphics.record_prefix)
        self.prefix.mkdir(exist_ok=True)
        (self.prefix / "pics").mkdir(exist_ok=True)
        (self.prefix / "movies").mkdir(exist_ok=True)

    def reset(self) -> None:
        self.identifier = uuid.uuid4()
        self.i = 0

    def record_frame(self, screen: Surface) -> None:
        if not Config().graphics.record:
            return

        pygame.image.save(
            screen,
            str(self.prefix / "pics" / f"pic_{self.identifier}_{self.i:04d}.png"),
        )
        self.i += 1

    def animate_game(self) -> None:
        if not Config().graphics.record:
            return

        import imageio_ffmpeg

        args = [
            imageio_ffmpeg.get_ffmpeg_exe(),
            "-framerate",
            f"{Config().graphics.record_fps}",
            "-i",
            str(self.prefix / "pics" / f"pic_{self.identifier}_%04d.png"),
            "-vcodec",
            "libx264",
            "-preset",
            "veryslow",
            "-crf",
            "17",
            "-tune",
            "animation",
            "-vf",
            f"fps={Config().graphics.record_fps},format=yuv420p",
            str(self.prefix / "movies" / f"game_{self.identifier}.mp4"),
        ]

        subprocess.run(args)

        for image in self.prefix.glob(f"pics/pic_{self.identifier}_*.png"):
            os.remove(str(image))



"""SnekEntity wrapping a Snek submission for the game world."""

import itertools
from collections import Counter
from collections.abc import Sequence
from dataclasses import dataclass, field

from typing_extensions import override

from sneks.engine.config.base import Config
from sneks.engine.core.action import Action
from sneks.engine.core.bearing import Bearing
from sneks.engine.core.cell import Cell
from sneks.engine.core.snek import Snek
from sneks.engine.entity.dynamic import DynamicEntity
from sneks.engine.manager.cell import CellManager
from sneks.engine.util import SwallowOutput


@dataclass(kw_only=True)
class SnekEntity(DynamicEntity):
    """Entity that manages the state of a Snek."""

    implementation: type[Snek]
    instance: Snek = field(init=False)
    current_food: frozenset[Cell] = field(init=False)

    @override
    def reset(
        self, initial_position: Cell | Sequence[Cell], initial_bearing: Bearing | None
    ) -> None:
        self.instance = self.implementation()
        super().reset(
            initial_position=initial_position, initial_bearing=initial_bearing
        )

    @override
    def before_step(self, occupied: frozenset[Cell], food: frozenset[Cell]) -> None:
        super().before_step(occupied=occupied, food=food)

        self.current_food = food

        if not self.ended:
            # set the state of the instance
            self.instance.bearing = self.bearing_manager.get_bearing()

            # build a grid around the head based on the vision range
            grid = {
                Cell(x, y)
                for x, y in itertools.product(
                    range(
                        self.head.x - Config().snek.vision_range,
                        self.head.x + Config().snek.vision_range,
                    ),
                    range(
                        self.head.y - Config().snek.vision_range,
                        self.head.y + Config().snek.vision_range,
                    ),
                )
            }

            # The sneks can only see obstacles within the vision range
            self.instance.occupied = frozenset(
                CellManager.get_relative_to(cell, self.head)
                for cell in grid.intersection(occupied)
                if cell.get_distance(self.head) < Config().snek.vision_range
            )

            # The snek can always know the locations of all the food
            self.instance.food = frozenset(
                CellManager.get_relative_to(cell, self.head) for cell in food
            )

            self.instance.body = list(
                CellManager.get_relative_to(cell, self.head)
                for cell in self.position_cells
            )

    @override
    def step(self, idle_allowed=True) -> None:
        super().step(idle_allowed=Config().snek.idle_allowed)

    @override
    def get_next_action(self) -> Action:
        with SwallowOutput(debug=Config().debug):
            next_action = self.instance.get_next_action()
        return next_action

    @override
    def after_step(self, occupied: Counter[Cell]) -> None:
        # check if snek should be ended
        if not self.ended:
            for cell in itertools.chain((self.head,), self.head_colliding_cells):
                if occupied[cell] > 1:
                    self.ended = True
                    self.ended_this_step = True
                    self.bearing_manager.update(Bearing(0, 0))
                    break
                if cell in self.current_food:
                    self.size += 1
        super().after_step(occupied=occupied)



"""Game execution and scoring runner."""

import multiprocessing
from concurrent.futures import ProcessPoolExecutor

from sneks.engine.config.base import Config
from sneks.engine.config.graphics import GraphicsConfig
from sneks.engine.scoring import NormalizedScore
from sneks.engine.world import World
from sneks.games.classic import world as classic_world
from sneks.games.classic.config import ClassicConfig


def run(
    world: type[World],
    config: Config,
    group_scores: bool = True,
) -> list[NormalizedScore]:
    if Config().profile:
        import cProfile
        import pstats

        pr = cProfile.Profile()
        pr.enable()

        result = _run(
            world=world,
            config=config,
            group_scores=group_scores,
        )

        pr.disable()
        stats = pstats.Stats(pr)
        stats.sort_stats("tottime").print_stats(20)

        return result
    else:
        return _run(
            world=world,
            config=config,
            group_scores=group_scores,
        )


def _run(
    world: type[World], config: Config, group_scores: bool
) -> list[NormalizedScore]:
    runs = 0
    scores: list[NormalizedScore] = []

    while runs < config.runs:
        current_world = world(config=config)
        current_world.run()
        score = current_world.score()
        score.sort(key=lambda s: s.total(criteria=world.criteria), reverse=True)
        if config.graphics.display:
            print(f"Run #{runs + 1}")
            NormalizedScore.pretty_print_headers(criteria=world.criteria)
            for s in score:
                s.pretty_print(criteria=world.criteria)
            print()
        scores.extend(score)
        runs += 1
        if runs % (config.runs / 20) == 0:
            print(f"{100 * runs / config.runs}% complete")

    if group_scores:
        grouped_scores = NormalizedScore.group(scores=scores, criteria=world.criteria)
        print("Aggregate scores:")
        NormalizedScore.pretty_print_headers(criteria=world.criteria)
        for s in grouped_scores:
            s.pretty_print(criteria=world.criteria)
        print()
        return grouped_scores
    return scores


def score_offline() -> None:
    batch_size = multiprocessing.cpu_count()
    runs = 0
    scores: list[NormalizedScore] = []
    previous_group: list[NormalizedScore] | None = None

    while True:
        with ProcessPoolExecutor() as executor:
            results = executor.map(score_offline_main, range(batch_size), timeout=20)
            completed = 0
            for result in results:
                try:
                    scores.extend(result)
                    completed += 1
                except TimeoutError:
                    pass

            runs += completed
            print("runs", runs)
            grouped = NormalizedScore.group(
                scores, criteria=classic_world.Classic.criteria
            )
            if isinstance(previous_group, list):
                change = 0
                for i in range(len(grouped)):
                    if previous_group[i].name != grouped[i].name:
                        change = 1
                    else:
                        previous_total = previous_group[i].total(
                            criteria=classic_world.Classic.criteria
                        )
                        grouped_total = grouped[i].total(
                            criteria=classic_world.Classic.criteria
                        )
                        change = max(
                            change,
                            abs(previous_total - grouped_total) / grouped_total
                            if grouped_total != 0
                            else 1.0,
                        )
                print("change %", change)
            previous_group = grouped
            NormalizedScore.pretty_print_headers(
                criteria=classic_world.Classic.criteria
            )
            for s in grouped:
                s.pretty_print(criteria=classic_world.Classic.criteria)


def score_offline_main(batch_id: int) -> list[NormalizedScore]:
    ClassicConfig(
        graphics=GraphicsConfig(
            display=False,
        )
    )
    scores = classic_world.main(group_scores=False, return_scores=True)
    if scores is None:
        raise ValueError
    return scores


if __name__ == "__main__":
    score_offline()



"""Space game mode with asteroids and multi-directional movement."""

import random
from typing import ClassVar, override

from sneks.engine import runner
from sneks.engine.core.bearing import Bearing
from sneks.engine.entity.static import StaticEntity
from sneks.engine.scoring import Criteria, NormalizedScore
from sneks.engine.world import World
from sneks.games.space.config import SpaceConfig


class Space(World):
    """Snek game with moving asteroids and respawning."""

    criteria: ClassVar[Criteria] = Criteria.CRASHES | Criteria.DISTANCE

    def __init__(self, config: SpaceConfig) -> None:
        super().__init__(config=config)

    @override
    def load_entities(self) -> None:
        super().load_entities()

        bearing_options = list(
            range(
                -SpaceConfig().snek.directional_speed_limit,
                SpaceConfig().snek.directional_speed_limit + 1,
            )
        )

        for _ in range(SpaceConfig().world.asteroid_count):
            up = random.choice(bearing_options)
            right = random.choice(bearing_options)
            if up == 0 and right == 0:
                # ensure all asteroids are moving
                increment = 1 if random.getrandbits(1) else -1
                if random.getrandbits(1):
                    up += increment
                else:
                    right += increment
            cluster = self.get_random_cluster(size=random.randint(3, 12))
            if not cluster:
                # Stop making asteroids if there's no longer room
                break
            self.register(
                StaticEntity(
                    name="Asteroid",
                    color=SpaceConfig().graphics.colors.asteroid,
                    initial_position=cluster,
                    initial_bearing=Bearing(right, up),
                ),
            )


def main(
    group_scores: bool = True, return_scores: bool = False
) -> list[NormalizedScore] | None:
    scores = runner.run(
        world=Space,
        config=SpaceConfig(),
        group_scores=group_scores,
    )
    if return_scores:
        return scores
    return None



"""Snek configuration for movement, vision, and respawn behavior."""

from dataclasses import dataclass

from sneks.engine.core.bearing import Bearing


@dataclass(frozen=True)
class SnekConfig:
    """Configuration for snek behavior and constraints."""

    vision_range: int = 20
    directional_speed_limit: int = 1
    combined_speed_limit: int = 1
    idle_allowed: bool = False
    initial_bearing: Bearing = Bearing(0, 1)
    respawn: bool = False



"""Scoring system with normalization and criteria flags."""

from dataclasses import dataclass
from enum import IntFlag, auto
from typing import Any


class Criteria(IntFlag):
    """Flags for which scoring metrics to include."""

    AGE = auto()
    CRASHES = auto()
    LENGTH = auto()
    DISTANCE = auto()


@dataclass(frozen=True)
class NormalizedScore:
    """Score normalized to 0-1 range for comparison."""

    name: str
    age: float
    crashes: float
    length: float
    distance: float
    raw: "Score"

    def __repr__(self):
        return (
            f"age': {self.age:.4f}, "
            f"crashes': {self.crashes:.4f}, "
            f"length': {self.length:.4f}, "
            f"distance': {self.distance:.4f}, "
            f"age: {self.raw.age:.4f}, "
            f"crashes: {self.raw.crashes:.4f}, "
            f"length: {self.raw.length:.4f}, "
            f"distance: {self.raw.distance:.4f}, "
            f"name: {self.name}"
        )

    def total(self, criteria: Criteria) -> float:
        total = 0.0
        if Criteria.AGE in criteria:
            total += self.age
        if Criteria.CRASHES in criteria:
            total += self.crashes
        if Criteria.LENGTH in criteria:
            total += self.length
        if Criteria.DISTANCE in criteria:
            total += self.distance
        return total

    @classmethod
    def from_dict(cls, dictionary: dict[str, Any]) -> "NormalizedScore":
        # Assume types are correct, expect a cast error otherwise
        return NormalizedScore(
            name=dictionary["name"],
            age=float(dictionary["age"]),
            crashes=float(dictionary["crashes"]),
            length=float(dictionary["length"]),
            distance=float(dictionary["distance"]),
            raw=Score(
                name=dictionary["name"],
                age=float(dictionary["raw"]["age"]),
                crashes=float(dictionary["raw"]["crashes"]),
                length=float(dictionary["raw"]["length"]),
                distance=float(dictionary["raw"]["distance"]),
            ),
        )

    @classmethod
    def pretty_print_headers(cls, criteria: Criteria) -> None:
        row = ["Total"]
        if Criteria.AGE in criteria:
            row.append("Age'")
        if Criteria.CRASHES in criteria:
            row.append("Crashes'")
        if Criteria.LENGTH in criteria:
            row.append("Length'")
        if Criteria.DISTANCE in criteria:
            row.append("Distance'")
        if Criteria.AGE in criteria:
            row.append("Age")
        if Criteria.CRASHES in criteria:
            row.append("Crashes")
        if Criteria.LENGTH in criteria:
            row.append("Length")
        if Criteria.DISTANCE in criteria:
            row.append("Distance")
        row.append("Name")

        print(" ".join(f"{col:<10}" for col in row))

    def pretty_print(self, criteria: Criteria) -> None:
        row: list[float | str] = [self.total(criteria=criteria)]
        if Criteria.AGE in criteria:
            row.append(self.age)
        if Criteria.CRASHES in criteria:
            row.append(self.crashes)
        if Criteria.LENGTH in criteria:
            row.append(self.length)
        if Criteria.DISTANCE in criteria:
            row.append(self.distance)
        if Criteria.AGE in criteria:
            row.append(self.raw.age)
        if Criteria.CRASHES in criteria:
            row.append(self.raw.crashes)
        if Criteria.LENGTH in criteria:
            row.append(self.raw.length)
        if Criteria.DISTANCE in criteria:
            row.append(self.raw.distance)
        row.append(self.name)
        print(
            " ".join(
                f"{float(col):<10.4f}" if not isinstance(col, str) else col
                for col in row
            )
        )

    @classmethod
    def group(
        cls, scores: list["NormalizedScore"], criteria: Criteria
    ) -> list["NormalizedScore"]:
        """Group the scores by name, and average them."""
        grouped: dict[str, list[NormalizedScore]] = {}
        for score in scores:
            if score.name not in grouped:
                grouped[score.name] = [score]
            else:
                grouped[score.name].append(score)
        final_scores = [
            NormalizedScore(
                name=group[0].name,
                age=sum(s.age for s in group) / len(group),
                crashes=sum(s.crashes for s in group) / len(group),
                length=sum(s.length for s in group) / len(group),
                distance=sum(s.distance for s in group) / len(group),
                raw=Score(
                    name=group[0].name,
                    age=sum(s.raw.age for s in group) / len(group),
                    crashes=sum(s.raw.crashes for s in group) / len(group),
                    length=sum(s.raw.length for s in group) / len(group),
                    distance=sum(s.raw.distance for s in group) / len(group),
                ),
            )
            for group in grouped.values()
        ]
        final_scores.sort(key=lambda s: s.total(criteria=criteria), reverse=True)
        return final_scores


@dataclass(frozen=True)
class Score:
    """Raw score values before normalization."""

    name: str
    age: float
    crashes: float
    length: float
    distance: float

    def normalize(self, min_score: "Score", max_score: "Score") -> NormalizedScore:
        return NormalizedScore(
            name=self.name,
            age=self.normalize_value(
                value=self.age, min_value=min_score.age, max_value=max_score.age
            ),
            crashes=self.normalize_value(
                value=self.crashes,
                min_value=min_score.crashes,
                max_value=max_score.crashes,
                invert=True,
            ),
            length=self.normalize_value(
                value=self.length,
                min_value=min_score.length,
                max_value=max_score.length,
            ),
            distance=self.normalize_value(
                value=self.distance,
                min_value=min_score.distance,
                max_value=max_score.distance,
            ),
            raw=self,
        )

    @classmethod
    def normalize_value(
        cls, value: float, min_value: float, max_value: float, invert: bool = False
    ) -> float:
        if invert:
            return (min_value - value) / max(1.0, (min_value - max_value))
        else:
            return (value - min_value) / max(1.0, (max_value - min_value))



"""Configuration for classic game mode."""

from dataclasses import dataclass

from sneks.engine.config.base import Config


@dataclass(frozen=True)
class ClassicConfig(Config):
    """Default configuration for classic snek."""

    pass



"""FoodEntity that sneks consume to grow."""

from collections import Counter
from dataclasses import dataclass
from typing import override

from sneks.engine.core.cell import Cell
from sneks.engine.entity.static import StaticEntity


@dataclass(kw_only=True)
class FoodEntity(StaticEntity):
    """A non-colliding entity that can trigger snek changes."""

    @override
    def after_step(self, occupied: Counter[Cell]) -> None:
        for cell in self.position_cells:
            if cell in occupied:
                self.ended = True



"""Asteroid entity for the space game mode."""

from dataclasses import dataclass

from sneks.engine.entity.static import StaticEntity


@dataclass(kw_only=True)
class Asteroid(StaticEntity):
    pass



"""Bearing primitive representing velocity in cells per step."""

from dataclasses import dataclass
from functools import cache, cached_property


@dataclass(frozen=True)
class Bearing:
    """
    Represents the directional speeds of a snek. The speeds are represented
    as integers to show the vertical and horizontal speeds in cells per
    game step.

    Sneks start with a bearing of ``(0, 0)``. For example, if the action from
    ``get_next_action()`` is ``Action.UP``, the snek will increase the speed vertically
    by one, so on the next game step the bearing will be ``(0, 1)``, and the snek will
    have moved one cell.
    """

    x: int  #:
    y: int  #:

    @cache  # type: ignore
    def __new__(cls, x: int, y: int) -> "Bearing":
        return super().__new__(cls)

    def __getnewargs__(self):
        return self.x, self.y

    @cached_property
    def _hash(self):
        return hash((self.x, self.y))

    def __eq__(self, other):
        return self._hash == other._hash

    def __hash__(self):
        return self._hash



import pytest

from sneks.engine.config.base import Config


def main(test_path: str | None = None, debug: bool = True) -> int:
    if test_path is not None:
        Config(debug=debug, registrar_prefix=test_path)
    else:
        Config(debug=debug)
    return pytest.main(["--pyargs", "sneks.engine.validator"])



"""Configuration for space game mode."""

from dataclasses import dataclass, field

from sneks.engine.config.base import Config
from sneks.engine.config.graphics import GraphicsConfig, TailDecorationBehavior
from sneks.engine.config.snek import SnekConfig
from sneks.engine.config.world import WorldConfig
from sneks.engine.core.bearing import Bearing


@dataclass(frozen=True)
class SpaceSnekConfig(SnekConfig):
    """Snek config for space mode with higher speed limits."""

    idle_allowed: bool = True
    directional_speed_limit: int = 2
    combined_speed_limit: int = 4
    initial_bearing: Bearing = Bearing(0, 0)
    respawn: bool = True


@dataclass(frozen=True)
class SpaceGraphicsConfig(GraphicsConfig):
    """Graphics config for space mode with decorative tails and stars."""

    tail_decoration: TailDecorationBehavior = (
        TailDecorationBehavior.POSITION_AND_DECORATIVE
    )
    draw_stars: bool = True


@dataclass(frozen=True)
class SpaceWorldConfig(WorldConfig):
    """World config for space mode with asteroids."""

    asteroid_count: int = 20
    food_count: int = 0


@dataclass(frozen=True)
class SpaceConfig(Config):
    """Configuration for space game mode."""

    title: str = "Sneks in SPACE"
    snek: SpaceSnekConfig = field(default_factory=SpaceSnekConfig)
    graphics: SpaceGraphicsConfig = field(default_factory=SpaceGraphicsConfig)
    world: SpaceWorldConfig = field(default_factory=SpaceWorldConfig)



"""Cell primitive representing a position on the game board."""

import math
from dataclasses import dataclass
from functools import cache, cached_property

from sneks.engine.config.base import Config
from sneks.engine.core.direction import Direction


@dataclass(frozen=True)
class Cell:
    """
    Represents a single cell on the game board. They can be used as reference points
    for getting neighboring cells, like:

    >>> cell = Cell(1, 1)
    >>> cell.get_left()
    Cell(0, 1)
    >>> cell.get_relative_neighbor(2, 3)
    Cell(3, 4)
    >>> cell.get_neighbor(Direction.DOWN)
    Cell(1, 0)

    Finally, it can be used to calculate distance between cells:

    >>> cell_a = Cell(0, 0)
    >>> cell_b = Cell(1, 0)
    >>> cell_a.get_distance(cell_b)
    1.0

    """

    x: int  #:
    y: int  #:

    @cache  # type: ignore
    def __new__(cls, x: int, y: int) -> "Cell":
        return super().__new__(cls)

    def __getnewargs__(self):
        return self.x, self.y

    @cached_property
    def _hash(self):
        # Determine the hash based on only positive indices,
        # so hash(Cell(-1,-1)) == hash(Cell(rows, columns))
        return hash((self.x % Config().world.columns, self.y % Config().world.rows))

    def __eq__(self, other):
        return self._hash == other._hash

    def __hash__(self):
        return self._hash

    @cache
    def get_relative_neighbor(self, x_offset: int, y_offset: int) -> "Cell":
        """
        Returns the cell with coordinates offset by the specified parameters.

        :param x_offset: the amount to offset this cell's x by
        :param y_offset: the amount to offset this cell's y by
        :return: the cell at ``(self.x + x_offset, self.y + y_offset)``
        """
        # We need to use integer modulus instead of floor to use the sign
        # of the numerator as opposed to the denominator in order to preserve
        # the relative direction of the Cell
        # Effectively: n - int(n / base) * base
        # Instead of:  n - floor(n / base) * base
        return Cell(
            int(math.fmod((self.x + x_offset), Config().world.columns)),
            int(math.fmod((self.y + y_offset), Config().world.rows)),
        )

    def get_neighbor(self, direction: Direction) -> "Cell":
        """
        Gets a Cell's neighbor in the specified direction.

        :param direction: the direction of the neighbor
        :return: cell in the specified direction
        """
        if direction is Direction.UP:
            return self.get_up()
        elif direction is Direction.DOWN:
            return self.get_down()
        elif direction is Direction.LEFT:
            return self.get_left()
        elif direction is Direction.RIGHT:
            return self.get_right()
        else:
            raise ValueError("direction not valid")

    def get_up(self) -> "Cell":
        """
        :return: cell in the up direction
        """
        return self.get_relative_neighbor(0, 1)

    def get_down(self) -> "Cell":
        """
        :return: cell in the down direction
        """
        return self.get_relative_neighbor(0, -1)

    def get_left(self) -> "Cell":
        """
        :return: cell in the left direction
        """
        return self.get_relative_neighbor(-1, 0)

    def get_right(self) -> "Cell":
        """
        :return: cell in the right direction
        """
        return self.get_relative_neighbor(1, 0)

    def get_distance(self, other: "Cell") -> float:
        """
        Gets the distance between this cell and another.

        :param other: cell to get distance to
        :return: distance between the two cells
        """

        # Get raw distance
        x_distance = abs(self.x - other.x)
        y_distance = abs(self.y - other.y)

        # Determine if it's closer to go the other way
        x_distance = (
            x_distance
            if x_distance < Config().world.columns / 2
            else Config().world.columns - x_distance
        )
        y_distance = (
            y_distance
            if y_distance < Config().world.rows / 2
            else Config().world.rows - y_distance
        )

        # Compute the distance
        return math.sqrt(x_distance**2 + y_distance**2)



"""BearingManager for mutable bearing state and velocity calculations."""

from collections.abc import Generator
from dataclasses import dataclass

from sneks.engine.config.base import Config
from sneks.engine.core.action import Action
from sneks.engine.core.bearing import Bearing
from sneks.engine.core.cell import Cell
from sneks.engine.core.direction import Direction


@dataclass
class BearingManager:
    """Manager for controlling the state of a bearing."""

    x: int = 0
    y: int = 0

    def get_bearing(self) -> Bearing:
        return Bearing(x=self.x, y=self.y)

    def update(self, bearing: Bearing) -> None:
        self.x = bearing.x
        self.y = bearing.y

    def apply(self, action: Action) -> None:
        """Apply the action to the bearing."""
        match action:
            case Action.UP:
                self.y = min(Config().snek.directional_speed_limit, self.y + 1)
            case Action.RIGHT:
                self.x = min(Config().snek.directional_speed_limit, self.x + 1)
            case Action.DOWN:
                self.y = max(-Config().snek.directional_speed_limit, self.y - 1)
            case Action.LEFT:
                self.x = max(-Config().snek.directional_speed_limit, self.x - 1)
        # Adjust the bearing based on the absolute speed limit
        while abs(self.x) + abs(self.y) > Config().snek.combined_speed_limit:
            # Remove from the vector not in the latest action
            # In theory, this should only remove a single unit per call,
            # so we're okay with ignoring the edge cases. If they get hit,
            # it's an indication of a configuration problem with the game
            match action:
                case Action.UP | Action.DOWN:
                    if self.x > 0:
                        self.x -= 1
                    elif self.x < 0:
                        self.x += 1
                    else:
                        break
                case Action.LEFT | Action.RIGHT:
                    if self.y > 0:
                        self.y -= 1
                    elif self.y < 0:
                        self.y += 1
                    else:
                        break
                case _:
                    break

    def get_distance(self) -> float:
        return Cell(0, 0).get_distance(Cell(self.x, self.y))

    def get_direction_components(self) -> Generator[Direction, None, None]:
        """
        Orders the directional components to make a path for the velocity
        """

        options = []
        if self.y > 0:
            options.append(Direction.UP)
        elif self.y < 0:
            options.append(Direction.DOWN)
        if self.x > 0:
            options.append(Direction.RIGHT)
        elif self.x < 0:
            options.append(Direction.LEFT)

        origin = Cell(0, 0)
        target = Cell(self.x, self.y)
        while origin != target:
            component = min(
                options,
                key=lambda direction: origin.get_neighbor(direction).get_distance(
                    target
                ),
            )
            origin = origin.get_neighbor(component)
            yield component



"""Graphics rendering using pygame."""

import hashlib
import itertools
import math
import os
import random
import struct
import sys
from collections import deque
from functools import cached_property

import pygame
from pygame import Surface

from sneks.engine.config.base import Config
from sneks.engine.config.graphics import GraphicsConfig, TailDecorationBehavior
from sneks.engine.core.cell import Cell
from sneks.engine.entity.base import Entity
from sneks.engine.entity.dynamic import DynamicEntity
from sneks.engine.entity.static import StaticEntity
from sneks.engine.manager.cell import CellManager
from sneks.engine.recorder import Recorder

os.environ["PYGAME_HIDE_SUPPORT_PROMPT"] = "1"


class Graphics:
    """Responsible for maintaining the graphics."""

    recorder: Recorder
    ended_head_splashes: deque[list[Cell]]

    def __init__(self) -> None:
        if not Config().graphics.display:
            return

        if Config().graphics.headless:
            os.environ["SDL_VIDEODRIVER"] = "dummy"

        pygame.init()
        self.screen = pygame.display.set_mode((self.width, self.height))
        pygame.display.set_caption(Config().title)
        self.recorder = Recorder()
        self.ended_head_splashes = deque(maxlen=10)

    def start(self) -> None:
        if not Config().graphics.display:
            return

    def paint(self, entities: list[Entity]) -> None:
        if not Config().graphics.display:
            return

        self._paint_board()
        if Config().graphics.draw_stars:
            self._paint_stars()

        ended_entities = []
        for entity in reversed(entities):
            match entity:
                case StaticEntity():
                    self.paint_static_entity(entity)
                case DynamicEntity():
                    if entity.ended_this_step:
                        ended_entities.append(entity.head)
                    self.paint_dynamic_entity(entity)
                case _:
                    raise RuntimeError

        self.ended_head_splashes.appendleft(ended_entities)
        self.paint_head_splashes()

        # check for exit conditions
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                sys.exit()

        # switch to the most recently drawn board
        pygame.display.flip()
        self.recorder.record_frame(self.screen)
        self.step_delay()

    def paint_static_entity(self, entity: StaticEntity) -> None:
        """Paint a blob with all cells connecting."""
        surface = Surface(
            (self.cell_size + self.padding, self.cell_size + self.padding)
        )
        surface.fill(entity.color)

        for cell in entity.position_cells:
            rect = self.get_cell_rect_from_surface(
                surface, cell, surface_has_padding=True
            )
            self.screen.blit(surface, rect)

    def paint_dynamic_entity(self, entity: DynamicEntity) -> None:
        """Paint a line using the visible deque."""
        cell_surface = Surface(
            (self.cell_size - self.padding, self.cell_size - self.padding)
        )
        horizontal_fill_surface = Surface(
            (self.padding * 2, self.cell_size - self.padding)
        )
        vertical_fill_surface = Surface(
            (self.cell_size - self.padding, self.padding * 2)
        )
        cell_surface.fill(entity.color)
        horizontal_fill_surface.fill(entity.color)
        vertical_fill_surface.fill(entity.color)

        # Determine which cells to draw
        iterables = []
        if TailDecorationBehavior.POSITION in Config().graphics.tail_decoration:
            iterables.append(entity.position_cells)
        if TailDecorationBehavior.COLLIDING in Config().graphics.tail_decoration:
            iterables.append(entity.head_colliding_cells)
            iterables.append(entity.tail_colliding_cells)
        if TailDecorationBehavior.DECORATIVE in Config().graphics.tail_decoration:
            iterables.append(entity.decorative_cells)

        color = entity.color
        previous: Cell | None = None
        start_gradiant_index = len(entity.position_cells)
        for i, cell in enumerate(itertools.chain(*iterables)):
            if i >= start_gradiant_index:
                color = tuple(int(c * 0.95) for c in color)
                cell_surface.fill(color)
                horizontal_fill_surface.fill(color)
                vertical_fill_surface.fill(color)

            # Blit cell
            rect = self.get_cell_rect_from_surface(surface=cell_surface, cell=cell)
            self.screen.blit(cell_surface, rect)

            # Blit padding
            self.blit_padding(
                cell=cell,
                previous=previous,
                horizontal_fill_surface=horizontal_fill_surface,
                vertical_fill_surface=vertical_fill_surface,
            )

            previous = cell

        # Blit head in distinct color
        if entity.ended:
            color = self.color_invalid
        else:
            color = struct.unpack(
                "BBB", hashlib.md5(struct.pack("BBB", *entity.color)).digest()[-3:]
            )
        cell_surface.fill(color)
        rect = self.get_cell_rect_from_surface(surface=cell_surface, cell=entity.head)
        self.screen.blit(cell_surface, rect)

    def paint_head_splashes(self) -> None:
        cell_surface = Surface(
            (self.cell_size - self.padding, self.cell_size - self.padding)
        )
        cell_surface.fill(self.color_invalid)
        for i, ended in enumerate(self.ended_head_splashes):
            for cell in ended:
                rect = self.get_cell_rect_from_surface(surface=cell_surface, cell=cell)
                self.screen.blit(cell_surface, rect)

                cell_surface.fill(tuple(int((0.85**i) * c) for c in self.color_invalid))

                splash_cells = [
                    CellManager.get_absolute_by_offset(
                        cell=cell, x_offset=x, y_offset=y
                    )
                    for x, y in itertools.product(
                        range(-int(i / 2), int(i / 2)), repeat=2
                    )
                ]
                for c in splash_cells:
                    rect = self.get_cell_rect_from_surface(surface=cell_surface, cell=c)
                    self.screen.blit(cell_surface, rect)

    def blit_padding(
        self,
        cell: Cell,
        previous: Cell | None,
        horizontal_fill_surface: Surface,
        vertical_fill_surface: Surface,
    ) -> None:
        # Blit padding
        if previous is not None:
            # Check if cell loops around the game board
            looped = abs(cell.x - previous.x) + abs(cell.y - previous.y) > 1

            relative = CellManager.get_relative_to(cell, previous)
            dx = min(1, max(relative.x, -1))
            dy = min(1, max(relative.y, -1))
            match (dx, dy):
                case (1, 0):
                    # cell is to the right, so fill left
                    rect = self.get_left_fill_from_surface(
                        surface=horizontal_fill_surface, cell=cell
                    )
                    self.screen.blit(horizontal_fill_surface, rect)
                    if looped:
                        # fill right of previous
                        rect = self.get_right_fill_from_surface(
                            surface=horizontal_fill_surface, cell=previous
                        )
                        self.screen.blit(horizontal_fill_surface, rect)
                case (-1, 0):
                    # cell is to the left, so fill right
                    rect = self.get_right_fill_from_surface(
                        surface=horizontal_fill_surface, cell=cell
                    )
                    self.screen.blit(horizontal_fill_surface, rect)
                    if looped:
                        # fill left of previous
                        rect = self.get_left_fill_from_surface(
                            surface=horizontal_fill_surface, cell=previous
                        )
                        self.screen.blit(horizontal_fill_surface, rect)
                case (0, 1):
                    # cell is up, so fill down
                    rect = self.get_down_fill_from_surface(
                        surface=vertical_fill_surface, cell=cell
                    )
                    self.screen.blit(vertical_fill_surface, rect)
                    if looped:
                        # fill up of previous
                        rect = self.get_up_fill_from_surface(
                            surface=vertical_fill_surface, cell=previous
                        )
                        self.screen.blit(vertical_fill_surface, rect)
                case (0, -1):
                    # cell is down, so fill up
                    rect = self.get_up_fill_from_surface(
                        surface=vertical_fill_surface, cell=cell
                    )
                    self.screen.blit(vertical_fill_surface, rect)
                    if looped:
                        # fill down of previous
                        rect = self.get_down_fill_from_surface(
                            surface=vertical_fill_surface, cell=previous
                        )
                        self.screen.blit(vertical_fill_surface, rect)

    def get_cell_rect_from_surface(
        self, surface: Surface, cell: Cell, surface_has_padding: bool = False
    ) -> pygame.Rect:
        offset = self.cell_size + (0 if surface_has_padding else self.padding)
        return surface.get_rect(
            top=offset
            + (Config().world.rows - cell.y - 1) * (self.cell_size + self.padding),
            left=offset + cell.x * (self.cell_size + self.padding),
        )

    def get_left_fill_from_surface(self, surface: Surface, cell: Cell) -> pygame.Rect:
        return surface.get_rect(
            top=self.cell_size
            + (Config().world.rows - cell.y - 1) * (self.cell_size + self.padding)
            + self.padding,
            left=self.cell_size
            + cell.x * (self.cell_size + self.padding)
            - self.padding,
        )

    def get_right_fill_from_surface(self, surface: Surface, cell: Cell) -> pygame.Rect:
        return surface.get_rect(
            top=self.cell_size
            + (Config().world.rows - cell.y - 1) * (self.cell_size + self.padding)
            + self.padding,
            left=self.cell_size
            + (cell.x + 1) * (self.cell_size + self.padding)
            - self.padding,
        )

    def get_down_fill_from_surface(self, surface: Surface, cell: Cell) -> pygame.Rect:
        return surface.get_rect(
            top=self.cell_size
            + (Config().world.rows - cell.y) * (self.cell_size + self.padding)
            - self.padding,
            left=self.cell_size
            + cell.x * (self.cell_size + self.padding)
            + self.padding,
        )

    def get_up_fill_from_surface(self, surface: Surface, cell: Cell) -> pygame.Rect:
        return surface.get_rect(
            top=self.cell_size
            + (Config().world.rows - cell.y - 1) * (self.cell_size + self.padding)
            - self.padding,
            left=self.cell_size
            + cell.x * (self.cell_size + self.padding)
            + self.padding,
        )

    def _paint_board(self) -> None:
        self.screen.fill(self.color_background)
        # borders
        top = (
            0,
            0,
            self.width,
            self.cell_size - self.padding,
        )
        bottom = (
            0,
            self.height - self.cell_size + self.padding,
            self.width,
            self.height,
        )
        left = (
            0,
            self.cell_size - self.padding,
            self.cell_size - self.padding,
            self.height,
        )
        right = (
            self.width - self.cell_size + self.padding,
            0,
            self.width,
            self.height,
        )
        for rect in (top, bottom, left, right):
            pygame.draw.rect(self.screen, self.color_border, rect)

    def _paint_stars(self) -> None:
        for center in self.stars:
            modifier = math.prod(center) % (256 - max(self.color_background))
            pygame.draw.circle(
                surface=self.screen,
                color=tuple(modifier + c for c in self.color_background),
                center=center,
                radius=1.0,
            )

    def end(self) -> None:
        if not Config().graphics.display:
            return

        self.end_delay()

    @property
    def rows(self) -> int:
        return Config().world.rows

    @property
    def columns(self) -> int:
        return Config().world.columns

    @property
    def cell_size(self) -> int:
        return Config().graphics.cell_size

    @property
    def padding(self) -> int:
        return Config().graphics.padding

    @property
    def color_border(self) -> tuple[int, int, int]:
        return Config().graphics.colors.border

    @property
    def color_background(self) -> tuple[int, int, int]:
        return Config().graphics.colors.background

    @property
    def color_invalid(self) -> tuple[int, int, int]:
        return Config().graphics.colors.invalid

    @property
    def height(self) -> int:
        return (2 + self.rows) * self.cell_size + self.rows * self.padding

    @property
    def width(self) -> int:
        return (2 + self.columns) * self.cell_size + self.columns * self.padding

    @cached_property
    def stars(self) -> list[tuple[int, int]]:
        return [
            (
                random.randint(
                    self.cell_size + self.padding,
                    self.width - self.cell_size - self.padding,
                ),
                random.randint(
                    self.cell_size + self.padding,
                    self.height - self.cell_size - self.padding,
                ),
            )
            for _ in range(self.rows + self.columns)
        ]

    def step_delay(self):
        if not Config().graphics.headless:
            if Config().graphics.step_keypress_wait:
                self.wait_for_keypress()
            pygame.time.delay(Config().graphics.step_delay)

    def end_delay(self):
        if not Config().graphics.headless:
            if Config().graphics.end_keypress_wait:
                self.wait_for_keypress()
            pygame.time.delay(Config().graphics.end_delay)

    @staticmethod
    def wait_for_keypress():
        while True:
            # allow the key to be held instead of waiting for each step
            if any(pygame.key.get_pressed()):
                break
            event = pygame.event.wait()
            if event.type == pygame.QUIT:
                pygame.quit()
                sys.exit()
            elif event.type == pygame.KEYDOWN:
                break


def generate_poster():
    """Draw the game board with nothing on it and save the image."""
    Config(
        graphics=GraphicsConfig(
            headless=True,
            record=True,
        ),
    )
    graphics = Graphics()
    graphics.paint([])



"""World base class managing the game loop, entities, and board state."""

import abc
import itertools
import logging
import random
from collections import Counter
from typing import ClassVar

from sneks.engine import registrar
from sneks.engine.config.base import Config
from sneks.engine.core.cell import Cell
from sneks.engine.core.direction import Direction
from sneks.engine.entity.base import Entity
from sneks.engine.entity.food import FoodEntity
from sneks.engine.entity.snek import SnekEntity
from sneks.engine.graphics import Graphics
from sneks.engine.scoring import Criteria, NormalizedScore, Score

logging.basicConfig(
    level=logging.INFO,
    format="{asctime} {levelname} {name}.{funcName}():{lineno} - {message}",
    style="{",
)
log = logging.getLogger(__name__)


class World(abc.ABC):
    """Responsible for maintaining the board and NPC entities."""

    cells: set[Cell]
    entities: list[Entity]
    config: Config
    graphics: Graphics
    halt_flag: bool = False
    steps: int = 0

    #: Worlds need to define their scoring criteria.
    criteria: ClassVar[Criteria] = Criteria(0)

    def __init__(self, config: Config) -> None:
        log.debug("begin")
        self.cells = {
            Cell(x, y)
            for x, y in itertools.product(
                range(config.world.columns), range(config.world.rows)
            )
        }
        self.entities = []
        self.config = config
        self.graphics = Graphics()

    def load_entities(self) -> None:
        sneks = registrar.get_submissions()
        sneks.sort(key=lambda s: s.name)
        color_index = 0
        color_index_delta = max(
            len(Config().graphics.colors.snek_options) // len(sneks), 1
        )
        for snek in sneks:
            starting_position = self.get_random_free_cell()
            if not starting_position:
                raise ValueError
            self.register(
                SnekEntity(
                    name=snek.name,
                    color=Config().graphics.colors.snek_options[color_index],
                    implementation=snek.snek,
                    initial_position=starting_position,
                    initial_bearing=Config().snek.initial_bearing,
                )
            )
            color_index = (color_index + color_index_delta) % len(
                Config().graphics.colors.snek_options
            )

        # Add initial food
        # Perform an optimization here for building the random free cells
        options = list(self.cells.difference(self.get_occupied_cells()))
        random.shuffle(options)
        for i in range(Config().world.food_count):
            if not options:
                break
            self.register(
                FoodEntity(
                    name="Food",
                    color=Config().graphics.colors.food,
                    initial_position=options.pop(),
                ),
            )
        log.debug("end")

    def register(self, entity: Entity) -> None:
        log.debug("begin")
        self.entities.append(entity)
        log.debug("end")

    def start(self) -> None:
        log.debug("begin")
        self.graphics.start()
        log.debug("end")

    def before_step(self) -> None:
        log.debug("begin")
        # Handle ended entities
        free_options = list(
            self.cells.difference(self.get_occupied_cells(include_food=True))
        )
        random.shuffle(free_options)
        for entity in self.entities:
            if not free_options:
                break
            if entity.ended:
                match entity:
                    case SnekEntity():
                        if Config().snek.respawn:
                            entity.ended = False
                            entity.reset(
                                initial_position=free_options.pop(),
                                initial_bearing=Config().snek.initial_bearing,
                            )
                    case FoodEntity():
                        entity.ended = False
                        entity.reset(
                            initial_position=free_options.pop(),
                            initial_bearing=None,
                        )
        # Set the state of the entity
        occupied = self.get_occupied_cells(include_food=False)
        food = self.get_food_cells()
        for entity in self.entities:
            entity.before_step(occupied=occupied, food=food)
        log.debug("end")

    def step(self) -> None:
        log.debug("begin")
        # Check the integrity of the engine
        if self.config is not Config():
            raise RuntimeError
        # Do the step
        for entity in self.entities:
            entity.step()
        log.debug("end")

    def after_step(self) -> None:
        log.debug("begin")
        # Evaluate consequences
        occupied = Counter()
        for entity in (e for e in self.entities if not isinstance(e, FoodEntity)):
            # need to count all the cells from an entity and the extra collision cells
            # first, update the position cells, ensuring duplicates are counted separately
            occupied.update(entity.position_cells)
            # next, update the collision cells, ensuring they're only counted at most once
            # from a particular entity
            collisions = (
                set(entity.head_colliding_cells) | set(entity.tail_colliding_cells)
            ) - set(entity.position_cells)
            occupied.update(collisions)
        for entity in self.entities:
            entity.after_step(occupied=occupied)
        self.graphics.paint(entities=self.entities)
        log.debug("end")

    def end(self) -> None:
        log.debug("begin")
        self.graphics.end()
        log.debug("end")

    def interrupt(self) -> None:
        log.debug("begin")
        self.halt_flag = True
        log.debug("end")

    def should_continue(self) -> bool:
        log.debug("begin")
        if self.halt_flag:
            return False
        if self.steps > Config().turn_limit:
            return False
        all_sneks_ended = all(
            s.ended for s in self.entities if isinstance(s, SnekEntity)
        )
        if not Config().snek.respawn and all_sneks_ended:
            return False
        return True

    def get_occupied_cells(self, include_food: bool = True) -> frozenset[Cell]:
        """Gets the set of occupied cells.

        This is determining a free cell to place a new entity, where
        food would be included in the set. It is also used for passing to the sneks
        when setting their state before a step, in which case food would be excluded.
        """
        return frozenset().union(
            *itertools.chain(
                entity.position_cells
                for entity in self.entities
                if include_food or not isinstance(entity, FoodEntity)
            )
        )

    def get_food_cells(self) -> frozenset[Cell]:
        return frozenset().union(
            *itertools.chain(
                entity.position_cells
                for entity in self.entities
                if isinstance(entity, FoodEntity)
            )
        )

    def get_random_free_cell(self) -> Cell | None:
        options = self.cells.difference(self.get_occupied_cells(include_food=True))
        if options:
            return random.choice(list(options))
        return None

    def get_random_cluster(self, size: int) -> list[Cell] | None:
        options = self.cells.difference(self.get_occupied_cells(include_food=True))
        if options:
            shuffled = list(options)
            random.shuffle(shuffled)
            for candidate in shuffled:
                attempts = 0
                potential = set()
                potential.add(candidate)

                while attempts < 100 and len(potential) < size:
                    attempts += 1
                    next_candidates = set()
                    for p in potential:
                        next_candidates.update(p.get_neighbor(d) for d in Direction)
                    next_candidates.intersection_update(options)
                    next_candidates.difference_update(potential)
                    if next_candidates:
                        shuffled_next_candidates = list(next_candidates)
                        random.shuffle(shuffled_next_candidates)
                        remaining_size = size - len(potential)
                        potential.update(shuffled_next_candidates[:remaining_size])

                if len(potential) == size:
                    return list(potential)
        return None

    def perform_step(self) -> None:
        self.before_step()
        self.step()
        self.after_step()
        self.steps += 1

    def run(self) -> None:
        if Config().profile:
            import cProfile
            import pstats

            pr = cProfile.Profile()
            pr.enable()

            self._run()

            pr.disable()
            stats = pstats.Stats(pr)
            stats.sort_stats("tottime").print_stats(20)
        else:
            self._run()

    def _run(self) -> None:
        log.debug("begin")
        self.load_entities()
        self.start()

        while self.should_continue():
            self.perform_step()

        self.end()
        log.debug("end")

    def score(self) -> list[NormalizedScore]:
        scores = []
        for entity in self.entities:
            if isinstance(entity, SnekEntity):
                scores.append(
                    Score(
                        name=entity.name,
                        age=entity.steps,
                        crashes=entity.count_ended,
                        length=entity.size,
                        distance=entity.distance,
                    )
                )
        min_score = Score(
            name="min",
            age=min(s.age for s in scores),
            crashes=max(s.crashes for s in scores),
            length=min(s.length for s in scores),
            distance=min(s.distance for s in scores),
        )
        max_score = Score(
            name="min",
            age=max(s.age for s in scores),
            crashes=min(s.crashes for s in scores),
            length=max(s.length for s in scores),
            distance=max(s.distance for s in scores),
        )
        return [s.normalize(min_score=min_score, max_score=max_score) for s in scores]



"""Configuration for toroidal game mode."""

from dataclasses import dataclass, field

from sneks.engine.config.base import Config
from sneks.engine.config.world import WorldConfig


@dataclass(frozen=True)
class ToroidalWorldConfig(WorldConfig):
    """World config for toroidal mode without food."""

    food_count: int = 0


@dataclass(frozen=True)
class ToroidalConfig(Config):
    """Configuration for toroidal game mode."""

    title: str = "Sneks on a Toroidal Plane"
    world: ToroidalWorldConfig = field(default_factory=ToroidalWorldConfig)
    turn_limit: int = 10000
